'''
Sample OpenLMI scripton for volume group management.
'''

from lmi.common import *
import time
from optparse import OptionParser
import logging


###############################################################################
# Scripton API
# - higher level functions to list, create, modify and delete volume groups

def get_vg(vgname=None, nm=None):
    """
    Return a LmiInstance representing volume group with given name.
    
    :param vgname: (``string``) Name of the volume group.
    """
    nm = get_nm(nm)
    vg = nm.LMI_VGStoragePool.first_instance(Key='ElementName', Value=vgname)
    if not vg:
        raise LMI_NotFoundException("Cannot find volume group '%s'" % (vgname))

    return vg

def get_vgs(nm=None):
    """
    Return a list LmiInstances of all volume groups.
    
    """
    nm = get_nm(nm)
    vgs = nm.LMI_VGStoragePool.instances()
    return vgs

def vg_rename(new_name, vgname=None, vgpool=None, nm=None):
    """
    Rename given volume group.
    :param new_name: (``string``) New name of the volume group.
    :param vgname: (``string``) Old name of the volume group. Either ``vgname``
    or ``vgpool`` must be specified.
    :param vgpool: (``LmiInstance``) Instance representing the volume group.
    Either ``vgname`` or ``vgpool`` must be specified.
    :param nm: (``LmiNameSpace``) CIM namespace with LMI Storage classes.
    The global one will be used when none is specified.
    """
    # Blivet does not support renaming, so it's just as example here...
    nm = get_nm(nm)

    if not vgpool:
        vgpool = get_vg(vgname)

    srv = _get_storage_service(nm)
    (ret, _out, _err) = srv.CreateOrModifyVG(ElementName=new_name, Pool=vgpool)
    if ret != 0:
        # TODO: we need a way how to translate 'ret' to value map name
        raise LMI_MethodFailed("Cannot rename volume group")
    # no return value

def vg_create(name, devices, extent_size=None, nm=None):
    """
    Create new volume group out of given devices and return it. The devices
    are automatically formatted as physical volumes.
    
    :param name: (``string``) Name of the volume group to create. If None,
    unique name will be generated.
    :param devices: (``list of LmiInstances of LMI_StorageExtents``) List of
    devices to add to the volume group.
    
    :returns: LmiInstance of the LMI_VGStoragePool.
    """
    nm = get_nm(nm)
    srv = _get_storage_service(nm)

    params = {
            'ElementName' : name,
            'InExtents' : devices
    }

    if extent_size:
        # we must create settings for the extent_size
        cap = nm.LMI_VGStorageCapabilities.first_instance()
        (ret, out, _err) = cap.CreateSetting()
        if ret != 0:
            raise LMI_MethodFailed(
                    "LMI_VGStorageCapabilities.CreateSetting failed to create "
                    "setting for the new volume group.")
        setting = out['newsetting'].to_instance()
        # TODO: do not import pywbem when ticket #123 if fixed
        import pywbem
        setting.ExtentSize = pywbem.Uint64(1024 * 1024)
        setting.push()
        params['Goal'] = setting

    (ret, out, _err) = srv.CreateOrModifyVG(**params)

    # TODO: wait for the job. lmishell ought to have a nicer way for this
    if out.has_key('Job'):
        job = out['Job'].to_instance()
        (ret, out, _err) = _wait_for_job(job, nm)
        # TODO: remove this dirty hack once Pegasus supports embedded instances
        out['Pool'] = devices[0].first_associator(
                ResultClass='LMI_VGStoragePool',
                AssocClass='LMI_VGAssociatedComponentExtent')
    if extent_size:
        # delete the setting
        setting.delete()

    if ret != 0:
        # TODO translate to an error
        raise LMI_MethodFailed("CreateOrModifyVG failed")
    return out['Pool']

def vg_destroy(vgname=None, vgpool=None, nm=None):
    """
    Destroy a volume group.
    
    :param vgname: (``string``) Old name of the volume group. Either ``vgname``
    or ``vgpool`` must be specified.
    :param vgpool: (``LmiInstance``) Instance representing the volume group.
    Either ``vgname`` or ``vgpool`` must be specified.
    :param nm: (``LmiNameSpace``) CIM namespace with LMI Storage classes.
    The global one will be used when none is specified.
    """
    nm = get_nm(nm)
    srv = _get_storage_service(nm)
    if not vgpool:
        vgpool = get_vg(vgname, nm)

    (ret, out, _err) = srv.DeleteVG(
            Pool=vgpool)
    # TODO: wait for the job. lmishell ought to have a nicer way for this
    if out.has_key('Job'):
        (ret, out, _err) = _wait_for_job(out['Job'].to_instance(), nm)
    if ret != 0:
        # TODO translate to an error
        raise LMI_MethodFailed("DeleteVG failed")

def extend(new_devices, vgname=None, vgpool=None, nm=None):
    """
    Add given block devices to a volume group.
    
    :param new_devices: (``list of LMIInstances of CIM_StorageExtents``) Devices
    to add to the volume group.
    :param vgname: (``string``) Old name of the volume group. Either ``vgname``
    or ``vgpool`` must be specified.
    :param vgpool: (``LmiInstance``) Instance representing the volume group.
    Either ``vgname`` or ``vgpool`` must be specified.
    :param nm: (``LmiNameSpace``) CIM namespace with LMI Storage classes.
    The global one will be used when none is specified.
    """
    nm = get_nm(nm)
    srv = _get_storage_service(nm)
    if not vgpool:
        vgpool = get_vg(vgname)

    # we must add existing PVs to the array of devices
    pvs = vgpool.associators(AssocClass='LMI_VGAssociatedComponentExtent')
    (ret, out, _err) = srv.CreateOrModifyVG(
            Pool=vgpool,
            InExtents=new_devices + pvs)
    # TODO: wait for the job. lmishell ought to have a nicer way for this
    if out.has_key('Job'):
        (ret, out, _err) = _wait_for_job(out['Job'].to_instance(), nm)
    if ret != 0:
        # TODO translate to an error
        raise LMI_MethodFailed("CreateOrModifyVG failed")

def vg_get_pvs(vgname=None, vgpool=None, nm=None):
    """
    Return list of physical volumes of given volume group.
    
    to add to the volume group.
    :param vgname: (``string``) Old name of the volume group. Either ``vgname``
    or ``vgpool`` must be specified.
    :param vgpool: (``LmiInstance``) Instance representing the volume group.
    Either ``vgname`` or ``vgpool`` must be specified.
    :param nm: (``LmiNameSpace``) CIM namespace with LMI Storage classes.
    The global one will be used when none is specified.
    
    :returns: ``list of LmiInstances of CIM_StorageExtents``
    """
    return vgpool.associators(AssocClass='LMI_VGAssociatedComponentExtent')

def vg_get_lvs(vgname=None, vgpool=None, nm=None):
    """
    Return list of logical volumes allocated from given volume group.
    
    to add to the volume group.
    :param vgname: (``string``) Old name of the volume group. Either ``vgname``
    or ``vgpool`` must be specified.
    :param vgpool: (``LmiInstance``) Instance representing the volume group.
    Either ``vgname`` or ``vgpool`` must be specified.
    :param nm: (``LmiNameSpace``) CIM namespace with LMI Storage classes.
    The global one will be used when none is specified.
    
    :returns: ``list of LmiInstances of CIM_StorageExtents``
    """
    return vgpool.associators(AssocClass='LMI_VGAssociatedComponentExtent')

def extend_with_empty_disks(vgname=None, vgpool=None, nm=None):
    """
    Find all devices with no partitions, create partition on them and
    add them to given volume group.
    
    This function is intended to extend given volume group with empty disks
    which were just plugged in.
    
    :param vgname: (``string``) Old name of the volume group. Either ``vgname``
    or ``vgpool`` must be specified.
    :param vgpool: (``LmiInstance``) Instance representing the volume group.
    Either ``vgname`` or ``vgpool`` must be specified.
    :param nm: (``LmiNameSpace``) CIM namespace with LMI Storage classes.
    The global one will be used when none is specified.
    """
    nm = get_nm(nm)
    if not vgpool:
        vgpool = get_vg(vgname)

    # find all primordial extents (=disks) with no partitions
    disks = nm.CIM_StorageExtent.instances(Key='Primordial', Value='True')
    new = []
    for disk in disks:
        if not disk.associators(ResultClass='CIM_GenericPartition'):
            new.append(disk)

    extend(new, vgpool=vgpool)

###############################################################################
# Command line utilities

def cmd_vgs(args, nm):
    """ Simulate 'vgs' command. """

    # parse the command line
    usage = "Usage: vgs [options] [vgname] [...]"
    parser = OptionParser(usage=usage)
    parser.add_option("-n", "--noheadings",
            action="store_true",
            dest="noheadings",
            default=False,
            help="Suppress the headings line that is normally the first line of"
                    " output.")

    (options, args) = parser.parse_args(args)
    if not options.noheadings:
        print "VG\t#PV\t#LV\tVSize\tVFree"

    if not args:
        # no other options - print all vgs
        vgs = get_vgs(nm)
    else:
        # print vgs specified on cmdline
        vgs = []
        for vgname in args:
            try:
                vg = get_vg(vgname, nm)
                vgs.append(vg)
            except LMI_NotFoundException:
                # just print error and continue
                logging.getLogger('lmi').error("Volume group %s not found"
                        % (vgname,))

    for vg in vgs:
        pvs = vg_get_pvs(vgpool=vg, nm=nm)
        lvs = vg_get_lvs(vgpool=vg, nm=nm)
        print "%s\t%d\t%d\t%d\t%d\t" % (
                vg.ElementName,
                len(pvs),
                len(lvs),
                vg.TotalManagedSpace,
                vg.RemainingManagedSpace)


def cmd_vgcreate(args, nm):
    """ Simulate 'vgcreate' command. """

    # parse the command line
    usage = "Usage: vgcreate [options] <vgname>  <pv> [...]"
    parser = OptionParser(usage=usage)
    parser.add_option('-s', "--physicalextentsize",
            action="store",
            type="int",
            dest="extent_size",
            help="physical extent size, in bytes.")
    # TODO: allow 'm' and 'k' suffixes to comfortably set extent size in
    # MiB and KiB

    (options, args) = parser.parse_args(args)
    if len(args) < 2:
        parser.error("Too few options specified.")

    vgname = args[0]
    pvs = []
    for pvname in args[1:]:
        pv = _get_device(pvname, nm)
        pvs.append(pv)
    vg_create(vgname, pvs, options.extent_size, nm)


def cmd_vgremove(args, nm):
    """ Simulate 'vgremove' command. """

    # parse the command line
    usage = "Usage: vgremove [options] <vgname> [...]"
    parser = OptionParser(usage=usage)

    (options, args) = parser.parse_args(args)
    if len(args) < 1:
        parser.error("Too few options specified.")

    for vgname in args:
        vg_destroy(vgname=vgname, nm=nm)

###############################################################################
# Private helper functions

def _get_device(name, nm):
    """ Find CIM_StorageExtent with given name."""
    # TODO: this method should be in some lmi.storage.devices
    extents = nm.wql('SELECT * FROM CIM_StorageExtent WHERE Name="%s"' % (name))
    if not extents:
        raise LMI_NotFoundException("Cannot find device '%s'" % (name))
    if len(extents) > 1:
        raise LMI_NotFoundException("Multiple devices with name '%s' found"
                % (name))
    return extents[0]


def _get_storage_service(nm):
    """
    Return instance of LMI_StorageConfigurationService.
    :param nm: Namespace of the OpenLMI storage classes.
    """
    # TODO: there must be some caching, we don't want to look up the service
    # every time.
    # It's constant, so it should be possible to turn CIMInstancePath to
    # LmiInstance...
    srv = nm.LMI_StorageConfigurationService.first_instance()
    return srv

def _wait_for_job(job, nm):
    """
    This is auxiliary function until lmishell can handle jobs natively.
    """
    while True:
        i = nm.LMI_StorageJob.first_instance(Key='InstanceID',
                Value=job.InstanceID)
        if i.JobState > 4:
            # TODO: Pegasus does not support embedded instances, fix once it's
            # done.
            return (0, {}, None)

            outparams = i.JobOutParameters
            ret = outparams.__ReturnValue
            out = []
            for prop in outparams.properties:
                # TODO: this won't work, see trac #118
                out[prop] = outparams.get_property(prop)
            return (ret, out, None)
        time.sleep(0.1)
